from libifstate.util import logger, IfStateLogging

from copy import deepcopy
import logging
import multiprocessing as mp
import os
import re
from setproctitle import setproctitle
import signal

class VrrpFifoProcess(mp.Process):
    '''
    Process for vrrp group/instance configuration.
    '''
    def __init__(self, vrrp_type, vrrp_name, ifstate, log_level):
        self.vrrp_type = vrrp_type
        self.vrrp_name = vrrp_name
        self.ifstate = ifstate
        self.log_level = log_level
        self.queue = mp.Queue()
        super().__init__(target=self.vrrp_worker, name=f'ifstate-vrrp-fifo|{vrrp_type}.{vrrp_name}', daemon=True)

    def vrrp_update(self, vrrp_state):
        self.queue.put(vrrp_state)

    def vrrp_worker(self):
        instance_title = "vrrp-{}-{}".format(self.vrrp_type, self.vrrp_name)
        setproctitle("ifstate-{}".format(instance_title))
        ifslog = IfStateLogging(self.log_level, action=instance_title, log_stderr=False)
        logger.info('worker spawned')
        while True:
            vrrp_state = self.queue.get()
            if vrrp_state is None:
                logger.info('terminating')
                return
            else:
                ifstate = deepcopy(self.ifstate)
                ifstate.apply(self.vrrp_type, self.vrrp_name, vrrp_state)

class VrrpStates():
    '''
    Tracks processes and states for vrrp groups/instances.
    '''
    def __init__(self, ifstate, log_level):
        self.ifstate = ifstate
        self.log_level = log_level
        self.processes = {}
        self.states = {}

    def update(self, vrrp_type, vrrp_name, vrrp_state):
        '''
        Updates the state for a group/instance. A new VrrpFifoProcess is spawned on-demand
        if a group/instance is called the first time.
        '''
        key = (vrrp_type, vrrp_name)
        if not key in self.processes:
            self.processes[key] = VrrpFifoProcess(vrrp_type, vrrp_name, self.ifstate, self.log_level)
            self.processes[key].start()

        self.states[key] = vrrp_state
        self.processes[key].vrrp_update(vrrp_state)

    def reconfigure(self, ifstate):
        '''
        Reconfigure all known groups/instances using their last known state by spawning
        new worker Processes.
        '''
        self.ifstate = ifstate
        for key, state in self.states.items():
            self.processes[key].vrrp_update(None)
            self.processes[key] = VrrpFifoProcess(*key, ifstate, self.log_level)
            self.processes[key].start()
            self.processes[key].vrrp_update(self.states[key])

def vrrp_fifo(args_fifo, ifs_config, log_level):
    vrrp_states = VrrpStates(ifs_config.ifs, log_level)
    ifs_config.set_callback(vrrp_states.reconfigure)

    signal.signal(signal.SIGHUP, ifs_config.sighup_handler)
    signal.signal(signal.SIGPIPE, signal.SIG_IGN)
    signal.signal(signal.SIGTERM, signal.SIG_IGN)

    pid_file = f"/run/libifstate/vrrp/{os.getpid()}.pid"
    try:
        os.makedirs("/run/libifstate/vrrp", exist_ok=True)

        with open(pid_file, "w", encoding="utf-8") as fh:
            fh.write(args_fifo)
    except IOError as err:
        logger.exception(f'failed to write pid file {pid_file}: {err}')

    try:
        status_pattern = re.compile(
            r'(group|instance) "([^"]+)" (unknown|fault|backup|master)( \d+)?$', re.IGNORECASE)

        with open(args_fifo) as fifo:
            logger.debug("entering fifo loop...")
            for line in fifo:
                m = status_pattern.match(line.strip())
                if m:
                    vrrp_type = m.group(1)
                    vrrp_name = m.group(2)
                    vrrp_state = m.group(3)

                    vrrp_states.update(vrrp_type, vrrp_name, vrrp_state)
                else:
                    logger.warning(f'failed to parse fifo input: {line.strip()}')
                mp.active_children()
    finally:
        try:
            os.remove(pid_file)
        except IOError:
            pass



from contextlib import contextmanager
import rlcompleter
import atexit
import code
import os
import readline

import pprint
from libifstate.netns import NetNameSpace

has_pygments = False
try:
    from pygments import highlight
    from pygments.lexers import PythonLexer
    from pygments.formatters import TerminalFormatter

    has_pygments = True
except:
    pass


def pygments_print(obj):
    formated = pprint.pformat(obj)
    print(highlight(formated, PythonLexer(), TerminalFormatter()), end="")


class IfStateConsole(code.InteractiveConsole):
    def __init__(self, locals={}, filename="<ifstate-shell>",
                 histfile=os.path.expanduser("~/.ifstate_history")):

        # use pygments for print() if available
        print_func = pprint.pprint
        if has_pygments:
            print_func = pygments_print

        ipr = NetNameSpace(None).ipr

        print("Links:")
        for link in ipr.get_links():
            print("  {:2}: {}".format(
                link.get('index'), link.get_attr('IFLA_IFNAME')))
        print("")


        print("""Symbols:
  ipr = pyroute2.IPRoute()
""")

        locals['ipr'] = ipr
        locals['pprint'] = print_func

        readline.set_completer(rlcompleter.Completer(locals).complete)
        code.InteractiveConsole.__init__(self, locals, filename)

        self.h_len = 0
        self.init_history(histfile)

    def init_history(self, histfile):
        """load history and enable tab completion"""

        readline.parse_and_bind("tab: complete")
        if hasattr(readline, "read_history_file"):
            try:
                readline.read_history_file(histfile)
                self.h_len = readline.get_current_history_length()
            except FileNotFoundError:
                try:
                    open(histfile, 'wb').close()
                except:
                    pass
            atexit.register(self.save_history, histfile)

    def save_history(self, histfile):
        """append new history entries"""

        new_h_len = readline.get_current_history_length()
        readline.set_history_length(1000)
        readline.append_history_file(new_h_len - self.h_len, histfile)



import libifstate.exception
from libifstate.log import logger, IfStateLogging
from pyroute2 import IPRoute, NetNS, netns

from pyroute2.netlink.rtnl.tcmsg import tcmsg
from pyroute2.netlink.rtnl import RTM_DELTFILTER, RTM_NEWNSID
from pyroute2.netlink.rtnl.nsidmsg import nsidmsg
from pyroute2.netlink import NLM_F_REQUEST
from pyroute2.netlink import NLM_F_ACK
from pyroute2.netlink import NLM_F_CREATE
from pyroute2.netlink import NLM_F_EXCL

try:
    # pyroute2 <0.6
    from pyroute2.ethtool.ioctl import SIOCETHTOOL
except ModuleNotFoundError:
    # pyroute2 >= 0.6
    from pr2modules.ethtool.ioctl import SIOCETHTOOL

import socket
import fcntl
import struct
import array
import struct
import typing

# ethtool helper
ETHTOOL_GDRVINFO = 0x00000003  # Get driver info
STRUCT_DRVINFO = struct.Struct(
    "I" +    # cmd
    "32s" +  # driver
    "32s" +  # version
    "32s" +  # fw_version
    "32s" +  # bus_info
    "32s" +  # reserved1
    "12s" +  # reserved2
    "I" +    # n_priv_flags
    "I" +    # n_stats
    "I" +    # testinfo_len
    "I" +    # eedump_len
    "I"      # regdump_len
)

ETHTOOL_GPERMADDR = 0x00000020  # Get permanent hardware address
L2_ADDRLENGTH = 6  # L2 address length

root_ipr = typing.NewType("IPRouteExt", IPRoute)

def filter_ifla_dump(showall, ifla, defaults, prefix="IFLA"):
    dump = {}

    for key, default_value in defaults.items():
        current_value = next(iter(ifla.get_attrs("_".join((prefix, key.upper())))), None)

        if current_value is not None:
            if showall or default_value is None or default_value != current_value:
                dump[key] = current_value

    return dump

class IPRouteExt(IPRoute):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.__sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def del_filter_by_info(self, index=0, handle=0, info=0, parent=0):
        msg = tcmsg()
        msg['index'] = index
        msg['handle'] = handle
        msg['info'] = info
        if parent != 0:
            msg['parent'] = parent

        return tuple(self.nlm_request(
            msg,
            msg_type=RTM_DELTFILTER,
            msg_flags=NLM_F_REQUEST |
            NLM_F_ACK
        ))

    def get_businfo(self, ifname):
        data = array.array("B", struct.pack(
            "I", ETHTOOL_GDRVINFO))
        data.extend(b'\x00' * (STRUCT_DRVINFO.size - len(data)))

        ifr = struct.pack('16sP', ifname.encode(
            "utf-8"), data.buffer_info()[0])

        try:
            r = fcntl.ioctl(self.__sock.fileno(), SIOCETHTOOL, ifr)
        except OSError:
            return None

        drvinfo = STRUCT_DRVINFO.unpack(data)

        return drvinfo[4].decode('ascii').split('\x00')[0]

    def get_permaddr(self, ifname):
        data = array.array("B", struct.pack(
            "II", ETHTOOL_GPERMADDR, L2_ADDRLENGTH))
        data.extend(b'\x00' * L2_ADDRLENGTH)

        ifr = struct.pack('16sP', ifname.encode(
            "utf-8"), data.buffer_info()[0])

        try:
            r = fcntl.ioctl(self.__sock.fileno(), SIOCETHTOOL, ifr)
        except OSError:
            return None

        l2addr = ":".join(format(x, "02x") for x in data[8:])
        if l2addr == "00:00:00:00:00:00":
            return None

        return l2addr

    def get_iface_by_businfo(self, businfo):
        for iface in iter(self.get_links()):
            ifname = iface.get_attr('IFLA_IFNAME')
            bi = self.get_businfo(ifname)

            if bi and bi == businfo:
                return iface['index']

    def get_iface_by_permaddr(self, permaddr):
        for iface in iter(self.get_links()):
            ifname = iface.get_attr('IFLA_IFNAME')
            addr = self.get_permaddr(ifname)

            if addr and addr == permaddr:
                return iface['index']

        return None

    def get_ifname_by_index(self, index):
        link = next(iter(self.get_links(index)), None)

        if link is None:
            return index

        return link.get_attr('IFLA_IFNAME')

    def set_netnsid(self, nsid=None, pid=None, fd=None):
        '''
        call pyroute2's set_netnsid if available or use
        fallback implementation for pyroute2 <=0.79
        '''
        if hasattr(super(), 'set_netnsid'):
            return super().set_netnsid(nsid, pid, fd)
        else:
            msg = nsidmsg()

            if nsid is None or nsid < 0:
                # kernel auto select
                msg['attrs'].append(('NETNSA_NSID', 4294967295))
            else:
                msg['attrs'].append(('NETNSA_NSID', nsid))

            if pid is not None:
                msg['attrs'].append(('NETNSA_PID', pid))

            if fd is not None:
                msg['attrs'].append(('NETNSA_FD', fd))

            return self.nlm_request(msg, RTM_NEWNSID, NLM_F_REQUEST | NLM_F_ACK)

    def get_link(self, *argv, **kwarg):
        '''
        Returns the first link info by wrapping a `get_links()` call and return
        `None` rather than raising any pyroute2 netlink exception on error.
        '''
        try:
            return next(iter(self.get_links(*argv, **kwarg)), None)
        except Exception as err:
            if not isinstance(err, netlinkerror_classes):
                raise

        return None

class NetNSExt(NetNS):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        try:
            netns.pushns(self.netns)
            self.__sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        finally:
            netns.popns()

    def del_filter_by_info(self, index=0, handle=0, info=0, parent=0):
        msg = tcmsg()
        msg['index'] = index
        msg['handle'] = handle
        msg['info'] = info
        if parent != 0:
            msg['parent'] = parent

        return tuple(self.nlm_request(
            msg,
            msg_type=RTM_DELTFILTER,
            msg_flags=NLM_F_REQUEST |
            NLM_F_ACK
        ))

    def get_businfo(self, ifname):
        data = array.array("B", struct.pack(
            "I", ETHTOOL_GDRVINFO))
        data.extend(b'\x00' * (STRUCT_DRVINFO.size - len(data)))

        ifr = struct.pack('16sP', ifname.encode(
            "utf-8"), data.buffer_info()[0])

        try:
            r = fcntl.ioctl(self.__sock.fileno(), SIOCETHTOOL, ifr)
        except OSError:
            return None

        drvinfo = STRUCT_DRVINFO.unpack(data)

        return drvinfo[4].decode('ascii').split('\x00')[0]

    def get_permaddr(self, ifname):
        data = array.array("B", struct.pack(
            "II", ETHTOOL_GPERMADDR, L2_ADDRLENGTH))
        data.extend(b'\x00' * L2_ADDRLENGTH)

        ifr = struct.pack('16sP', ifname.encode(
            "utf-8"), data.buffer_info()[0])

        try:
            r = fcntl.ioctl(self.__sock.fileno(), SIOCETHTOOL, ifr)
        except OSError:
            return None

        l2addr = ":".join(format(x, "02x") for x in data[8:])
        if l2addr == "00:00:00:00:00:00":
            return None

        return l2addr

    def get_iface_by_businfo(self, businfo):
        for iface in iter(self.get_links()):
            ifname = iface.get_attr('IFLA_IFNAME')
            bi = self.get_businfo(ifname)

            if bi and bi == businfo:
                return iface['index']

    def get_iface_by_permaddr(self, permaddr):
        for iface in iter(self.get_links()):
            ifname = iface.get_attr('IFLA_IFNAME')
            addr = self.get_permaddr(ifname)

            if addr and addr == permaddr:
                return iface['index']

        return None

    def get_ifname_by_index(self, index):
        link = next(iter(self.get_links(index)), None)

        if link is None:
            return index

        return link.get_attr('IFLA_IFNAME')

    def set_netnsid(self, nsid=None, pid=None, fd=None):
        '''
        call pyroute2's set_netnsid if available or use
        fallback implementation for pyroute2 <=0.79
        '''
        if hasattr(super(), 'set_netnsid'):
            return super().set_netnsid(nsid, pid, fd)
        else:
            msg = nsidmsg()

            if nsid is None or nsid < 0:
                # kernel auto select
                msg['attrs'].append(('NETNSA_NSID', 4294967295))
            else:
                msg['attrs'].append(('NETNSA_NSID', nsid))

            if pid is not None:
                msg['attrs'].append(('NETNSA_PID', pid))

            if fd is not None:
                msg['attrs'].append(('NETNSA_FD', fd))

            return self.nlm_request(msg, RTM_NEWNSID, NLM_F_REQUEST | NLM_F_ACK)

    def get_link(self, *argv, **kwarg):
        '''
        Returns the first link info by wrapping a `get_links()` call and return
        `None` rather than raising any pyroute2 netlink exception on error.
        '''
        try:
            return next(iter(self.get_links(*argv, **kwarg)), None)
        except Exception as err:
            if not isinstance(err, libifstate.exception.netlinkerror_classes):
                raise

        return None

class LinkDependency:
    def __init__(self, ifname, netns):
        self.ifname = ifname
        self.netns = netns

    def __hash__(self):
        return hash((self.ifname, self.netns))

    def __eq__(self, other):
        return (self.ifname, self.netns) == (other.ifname, other.netns)

    def __lt__(self, obj):
        if self.netns is None:
            if obj.netns is not None:
                return True
        elif obj.netns is None:
            return False

        if self.netns != obj.netns:
            return self.netns < obj.netns

        if self.ifname == obj.ifname:
            return False

        if self.ifname == 'lo':
            return True

        if obj.ifname == 'lo':
            return False

        return self.ifname < obj.ifname

    def __ne__(self, other):
        return not(self == other)

    def __str__(self):
        if self.netns is None:
            return "{}".format(self.ifname)
        else:
            return "{}[netns={}]".format(self.ifname, self.netns)


root_ipr = IPRouteExt()



from libifstate.util import logger, IfStateLogging
from libifstate.bpf.map import BPF_Map
from libifstate.bpf.ctypes import *
import os
import pyroute2.netns
import shutil


BPF_PROG_TYPE_XDP = 6

bpfs_ifstate_dir = '/sys/fs/bpf/ifstate'


class BPF():
    def __init__(self, netns):
        self.netns = netns
        self.bpf_fds = {}
        self.bpf_progs = {}
        self.bpf_tags = {}

    def add(self, name, config):
        self.bpf_progs[name] = config

    def apply(self, do_apply):
        # prepare pinning directories
        if self.netns.netns is None:
            progs_pin_dir = "{}/progs".format(bpfs_ifstate_dir)
            maps_pin_dir = "{}/maps".format(bpfs_ifstate_dir)
        else:
            progs_pin_dir = "{}/netns/{}/progs".format(bpfs_ifstate_dir, self.netns.netns)
            maps_pin_dir = "{}/netns/{}/maps".format(bpfs_ifstate_dir, self.netns.netns)

        pinning = False
        if os.path.ismount('/sys/fs/bpf'):
            logger.debug('bpfs is mounted, enable pinning')
            pinning = True

            if do_apply:
                os.makedirs(progs_pin_dir, exist_ok=True)
                os.makedirs(maps_pin_dir, exist_ok=True)
        else:
            logger.debug('bpfs is not mounted, skipping pinning')

        if self.netns.netns is not None:
            pyroute2.netns.pushns(self.netns.netns)

        try:
            # load (and pin) bpf programs
            for name, config in self.bpf_progs.items():
                log_str = name
                if self.netns.netns is not None:
                    log_str += "[netns={}]".format(self.netns.netns)

                current_prog_tag = None
                current_prog_fd = -1
                self.bpf_tags[name] = None
                self.bpf_fds[name] = -1

                prog_pin_filename = "{}/{}".format(progs_pin_dir, name)
                if os.path.isfile(prog_pin_filename):
                    current_prog_fd = libbpf.bpf_obj_get(
                        os.fsencode(prog_pin_filename))

                    if current_prog_fd < 0:
                        logger.warning('could not get current BPF obj fd for {}: {}'.format(
                            log_str, os.strerror(-current_prog_fd)))
                    else:
                        self.bpf_fds[name] = current_prog_fd

                        current_prog_info = struct_bpf_prog_info()
                        current_prog_size = ctypes.c_uint(
                            ctypes.sizeof(current_prog_info))
                        rc = libbpf.bpf_obj_get_info_by_fd(current_prog_fd, ctypes.byref(
                            current_prog_info), ctypes.byref(current_prog_size))
                        if rc == 0:
                            current_prog_tag = bytes(current_prog_info.tag).hex()
                            self.bpf_tags[name] = current_prog_tag

                            logger.debug('current prog tag: {}'.format(current_prog_tag), extra={
                                'iface': log_str})
                        else:
                            logger.warning('could not get current BPF obj info for {}: {}'.format(
                                log_str, os.strerror(-rc)))

                # load new BPF prog
                new_prog_tag = None
                new_prog_fd = None

                new_obj = libbpf.bpf_object__open_file(
                    os.fsencode(config["object"]),
                    None,
                )
                if not new_obj:
                    logger.warning('BPF open on {} failed: {}'.format(
                        log_str, os.strerror(ctypes.get_errno())))
                    continue

                logger.debug('loaded object: {}'.format(libbpf.bpf_object__name(new_obj).decode('ascii')), extra={
                    'iface': log_str})

                rc = libbpf.bpf_object__load(new_obj)
                if rc < 0:
                    logger.warning('BPF load on {} failed: {}'.format(
                        log_str, os.strerror(-rc)))
                    continue

                new_prog = libbpf.bpf_object__next_program(new_obj, None)
                while new_prog:
                    section = libbpf.bpf_program__section_name(
                        new_prog).decode('ascii')

                    logger.debug('  section: {}'.format(section), extra={
                        'iface': log_str})

                    if section == config['section']:
                        break
                    prog = libbpf.bpf_object__next_program(new_obj, new_prog)

                if not new_prog:
                    logger.warning('BPF section {} on {} not found'.format(
                        config['section'], log_str))
                    continue

                # get prog fd
                new_prog_fd = libbpf.bpf_program__fd(new_prog)
                if not new_prog_fd:
                    logger.warning('BPF failed to get prog fd on {}'.format(
                        log_str))
                    continue

                # get new prog tag
                info = struct_bpf_prog_info()
                sz = ctypes.c_uint(ctypes.sizeof(info))
                rc = libbpf.bpf_obj_get_info_by_fd(
                    new_prog_fd, ctypes.byref(info), ctypes.byref(sz))
                if rc == 0:
                    new_prog_tag = bytes(info.tag).hex()

                    logger.debug('new prog tag: {}'.format(new_prog_tag), extra={
                        'iface': log_str})

                    if current_prog_tag != new_prog_tag:
                        self.bpf_fds[name] = new_prog_fd
                        self.bpf_tags[name] = new_prog_tag

                        if os.path.isfile(prog_pin_filename):
                            os.unlink(prog_pin_filename)
                        libbpf.bpf_program__pin(
                            new_prog,
                            os.fsencode(prog_pin_filename))

                        maps_path = "{}/{}".format(maps_pin_dir, name)

                        # unbind any orphan maps
                        if os.path.isdir(maps_path):
                            shutil.rmtree(maps_path)

                        # create a new maps directory
                        os.makedirs(maps_path, exist_ok=True)

                        # pin  maps
                        logger.debug('pin maps at: {}'.format(maps_path), extra={
                            'iface': log_str})
                        libbpf.bpf_object__pin_maps(
                            new_obj,
                            os.fsencode(maps_path))

                        logger.log_change(log_str)
                    else:
                        logger.log_ok(log_str)

                    # update maps if appropriate
                    if config.get('maps'):
                        for map_name, map_entries in config['maps'].items():
                            map_filename = "{}/{}/{}".format(maps_pin_dir, name, map_name)

                            try:
                                map_instance = BPF_Map(map_filename)
                            except OSError as ex:
                                logger.warning(
                                    "opening bpf map {} failed: {}".format(map_name, ex))
                                continue

                            # TODO: support to manage map entries
                else:
                    logger.warning(
                        'BPF failed to get prog info on {}'.format(log_str))
        finally:
            if self.netns.netns is not None:
                pyroute2.netns.popns()

    def get_bpf(self, name):
        return (self.bpf_fds.get(name, -1), self.bpf_tags.get(name))



from libifstate.exception import LinkDuplicate
from libifstate.link.base import ethtool_path, Link
from libifstate.address import Addresses
from libifstate.fdb import FDB
from libifstate.neighbour import Neighbours
from libifstate.routing import Tables, Rules, RTLookups
from libifstate.parser import Parser
from libifstate.tc import TC
from libifstate.exception import netlinkerror_classes, NetNSNotRoot
import bisect
import os
import pyroute2

from pyroute2.netlink.rtnl.ifaddrmsg import IFA_F_PERMANENT
try:
    from libifstate.wireguard import WireGuard
except ModuleNotFoundError:
    # ignore missing plugin
    pass
except Exception as err:
    # ignore plugin failure if kmod is not loaded
    if not isinstance(err, netlinkerror_classes):
        raise

try:
    from libifstate.bpf import libbpf, BPF

    try:
        from libifstate.xdp import XDP
    except ModuleNotFoundError:
        # ignore missing plugin
        pass
except ModuleNotFoundError:
    # ignore missing plugin
    pass

from libifstate.netns import NetNameSpace, prepare_netns, LinkRegistry, get_netns_instances
from libifstate.util import logger, IfStateLogging, LinkDependency
from libifstate.exception import FeatureMissingError, LinkCircularLinked, LinkNoConfigFound, ParserValidationError
from ipaddress import ip_network, ip_interface
from jsonschema import validate, ValidationError, FormatChecker
from copy import deepcopy
import os
import pkgutil
import re
import secrets
import json
import errno
import logging

__version__ = "1.11.8"


class IfState():
    def __init__(self):
        logger.debug('IfState {}'.format(__version__))

        self.namespaces = None
        self.root_netns = NetNameSpace(None)
        self.defaults = []
        self.ignore = {}
        self.features = {
            'brport': True,
            'link': True,
            'sysctl': os.access('/proc/sys/net', os.R_OK),
            'ethtool': not ethtool_path is None,
            'tc': True,
            'wireguard': not globals().get("WireGuard") is None,
            'bpf': not globals().get("libbpf") is None,
            'xdp': not globals().get("XDP") is None,
        }

        logger.debug('{}'.format(' '.join(sorted(
            [x for x, y in self.features.items() if y]))), extra={'iface': 'features'})

        if self.features['bpf']:
            if logger.level != logging.DEBUG:
                # BPF: disable libbpf stderr output
                libbpf.libbpf_set_print(0)

    def update(self, ifstates, soft_schema):
        # check config schema
        schema = json.loads(pkgutil.get_data(
            "libifstate", "../schema/ifstate.conf.schema.json"))
        try:
            validate(ifstates, schema, format_checker=FormatChecker())
        except ValidationError as ex:
            if len(ex.path) > 0:
                path = ["$"]
                for i, p in enumerate(ex.absolute_path):
                    if type(p) == int:
                        path.append("[{}]".format(p))
                    else:
                        path.append(".")
                        path.append(p)

                detail = "{}: {}".format("".join(path), ex.message)
            else:
                detail = ex.message
            if soft_schema:
                logger.error("Config validation failed for {}".format(detail))
            else:
                raise ParserValidationError(detail)

        # add interface defaults
        if 'defaults' in ifstates:
            self.defaults = ifstates['defaults']

        # add ignore list items
        self.ignore.update(ifstates['ignore'])
        self.ipaddr_ignore = set()
        for ip in self.ignore.get('ipaddr', []):
            self.ipaddr_ignore.add(ip_network(ip))

        # save cshaper profiles
        self.cshaper_profiles = ifstates['cshaper']

        # build link registry over all named netns
        self.link_registry = LinkRegistry(self.ignore.get('ifname', []), self.root_netns)

        self._update(self.root_netns, ifstates)
        if 'namespaces' in ifstates:
            # require to called from the root netns
            try:
                # assume init runs in the root netns
                if os.readlink('/proc/1/ns/net') != os.readlink(f'/proc/{os.getpid()}/ns/net'):
                    raise NetNSNotRoot()
            except OSError as ex:
                logger.debug(f'root netns test: {ex}')
                pass
            self.namespaces = {}
            self.new_namespaces = []
            for netns_name, netns_ifstates in ifstates['namespaces'].items():
                is_new = netns_name not in pyroute2.netns.listnetns()
                self.namespaces[netns_name] = NetNameSpace(netns_name)
                if is_new:
                    self.new_namespaces.append(netns_name)
                    self.link_registry.inventory_netns(self.namespaces[netns_name])
                self._update(self.namespaces[netns_name], netns_ifstates)

    def _update(self, netns, ifstates):
        # parse options
        if 'options' in ifstates:
            # parse global sysctl settings
            if 'sysctl' in ifstates['options']:
                for proto in  ifstates['options']['sysctl'].keys():
                    if proto in ['all', 'default']:
                        netns.sysctl.add(
                            proto, ifstates['options']['sysctl'][proto])
                    else:
                        netns.sysctl.add_global(
                            proto, ifstates['options']['sysctl'][proto])

        # load BPF programs
        if 'bpf' in ifstates:
            if not self.features['bpf']:
                raise FeatureMissingError("bpf")

            if netns.bpf_progs is None:
                netns.bpf_progs = BPF(netns)
            for name, config in ifstates['bpf'].items():
                netns.bpf_progs.add(name, config)

        # add interfaces from config
        for ifstate in ifstates['interfaces']:
            name = ifstate['name']
            kind = ifstate['link']['kind']
            defaults = self.get_defaults(
                ifname=name,
                kind=kind)

            if name in netns.links:
                raise LinkDuplicate()

            # prepare ethtool settings
            if name != 'lo':
                ethtool = {}
                if 'ethtool' in defaults:
                    ethtool.update(defaults['ethtool'])
                for k, v in ifstate.get('ethtool', {}).items():
                    if k in ethtool:
                        ethtool[k].update(v)
                    else:
                        ethtool[k] = v
                if not ethtool:
                    ethtool = None
            else:
                ethtool = None

            link = {}
            if 'link' in defaults:
                link.update(defaults['link'])
            if 'link' in ifstate:
                link.update(ifstate['link'])
            if link:
                netns.links[name] = Link(self,
                    netns, name, link, ethtool, ifstate.get('vrrp'), ifstate.get('brport'))
            else:
                netns.links[name] = None

            if 'addresses' in ifstate:
                netns.addresses[name] = Addresses(netns, name, ifstate['addresses'])
            elif defaults.get('clear_addresses', False):
                netns.addresses[name] = Addresses(netns, name, [])

            if 'fdb' in ifstate:
                netns.fdb[name] = FDB(netns, name, ifstate['fdb'])
            elif defaults.get('clear_fdb', False):
                netns.fdb[name] = FDB(netns, name, [])

            if 'neighbours' in ifstate:
                netns.neighbours[name] = Neighbours(netns, name, ifstate['neighbours'])
            elif defaults.get('clear_neighbours', False):
                netns.neighbours[name] = Neighbours(netns, name, [])

            if 'vrrp' in ifstate:
                ktype = ifstate['vrrp']['type']
                kname = ifstate['vrrp']['name']
                kstates = ifstate['vrrp']['states']
                if not kname in netns.vrrp[ktype]:
                    netns.vrrp[ktype][kname] = {}
                for kstate in kstates:
                    if not kstate in netns.vrrp[ktype][kname]:
                        netns.vrrp[ktype][kname][kstate] = []
                    netns.vrrp[ktype][kname][kstate].append(name)
                netns.vrrp['links'].append(name)

            if 'sysctl' in ifstate:
                netns.sysctl.add(name, ifstate['sysctl'])

            if 'cshaper' in ifstate:
                profile_name = ifstate['cshaper'].get(
                    'profile', 'default')
                logger.debug('cshaper profile {} enabled'.format(profile_name),
                             extra={'iface': name, 'netns': netns})
                cshaper_profile = deepcopy(self.cshaper_profiles[profile_name])

                # ingress
                ifb_name = re.sub(
                    cshaper_profile['ingress_ifname']['search'], cshaper_profile['ingress_ifname']['replace'], name)
                logger.debug('cshaper ifb name {}'.format(ifb_name),
                             extra={'iface': name, 'netns': netns})

                ifb_state = {
                    'name': ifb_name,
                    'link': {
                        'state': 'up',
                        'kind': 'ifb',
                    },
                    'tc': {
                        'qdisc': cshaper_profile['ingress_qdisc'],
                    }
                }
                ifb_state['tc']['qdisc']['bandwidth'] = ifstate['cshaper'].get(
                    'ingress', 'unlimited')

                if 'vrrp' in ifstate:
                    ifb_state['vrrp'] = ifstate['vrrp']

                ifstates['interfaces'].append(ifb_state)

                # egress
                if 'tc' in ifstate:
                    logger.warning(
                        'cshaper settings replaces tc settings', extra={'iface': name, 'netns': netns})

                ifstate['tc'] = {
                    'ingress': True,
                    'qdisc': cshaper_profile['egress_qdisc'],
                    'filter': [
                        {
                            'kind': 'matchall',
                            'parent': 'ffff:',
                            'action': [
                                {
                                    'kind': 'mirred',
                                    'direction': 'egress',
                                    'action': 'redirect',
                                    'dev': ifb_name,
                                }
                            ]
                        }

                    ]
                }

                ifstate['tc']['qdisc']['bandwidth'] = ifstate['cshaper'].get(
                    'egress', 'unlimited')

                del ifstate['cshaper']

            if 'tc' in ifstate:
                netns.tc[name] = TC(
                    netns, name, ifstate['tc'])
            elif defaults.get('clear_tc', False):
                netns.tc[name] = TC(
                    netns, name, {
                        'ingress': False,
                        'filter': [],
                        'qdisc': None
                    })

            if 'wireguard' in ifstate:
                if not self.features['wireguard']:
                    raise FeatureMissingError("wireguard")

                netns.wireguard[name] = WireGuard(netns, name, ifstate['wireguard'])

            if 'xdp' in ifstate:
                if not self.features['xdp']:
                    raise FeatureMissingError("xdp")

                netns.xdp[name] = XDP(netns, name, ifstate['xdp'])

        # add routing from config
        if 'routing' in ifstates:
            if 'routes' in ifstates['routing']:
                if netns.tables is None:
                    netns.tables = Tables(netns)
                for route in ifstates['routing']['routes']:
                    netns.tables.add(route)

            if 'rules' in ifstates['routing']:
                if netns.rules is None:
                    netns.rules = Rules(netns)
                for rule in ifstates['routing']['rules']:
                    netns.rules.add(rule)

    def apply(self, vrrp_type=None, vrrp_name=None, vrrp_state=None):
        self._apply(True, vrrp_type, vrrp_name, vrrp_state)

    def check(self, vrrp_type=None, vrrp_name=None, vrrp_state=None):
        self._apply(False, vrrp_type, vrrp_name, vrrp_state)

    def free_registry_item(self, do_apply, item):
        ifname = item.attributes['ifname']

        log_str = ifname
        if item.netns.netns is not None:
            log_str += "[netns={}]".format(item.netns.netns)

        if item.attributes['kind'] != 'physical':
            # remove virtual interface
            logger.log_del(log_str)
            if do_apply:
                try:
                    item.netns.ipr.link('set', index=item.attributes['index'], state='down')
                    item.netns.ipr.link('del', index=item.attributes['index'])
                except Exception as err:
                    if not isinstance(err, netlinkerror_classes):
                         raise
                    # ignore if the link is already gone, this might happen
                    # when removing veth link peers
                    if err.code != errno.ENODEV:
                        logger.warning('removing link {} failed: {}'.format(
                            ifname, err.args[1]), extra={'netns': item.netns})
            return True
        else:
            # shutdown physical interfaces
            if item.state == 'down':
                logger.log_ok(log_str, 'orphan')
            else:
                logger.log_del(log_str, 'orphan')
                if do_apply:
                    try:
                        item.netns.ipr.link('set', index=item.attributes['index'], state='down')
                    except Exception as err:
                        if not isinstance(err, netlinkerror_classes):
                            raise
                        logger.warning('updating link {} failed: {}'.format(
                            ifname, err.args[1]), extra={'netns': item.netns})
            return False

    def _dependencies(self, netns):
        deps = {}
        for ifname, link in netns.links.items():
            deps[LinkDependency(ifname, netns.netns)] = link.depends()

        for ifname, tc in netns.tc.items():
            for fltr in tc.tc.get('filter', []):
                for action in fltr.get('action', []):
                    if 'dev' in action:
                        link = LinkDependency(ifname, netns.netns)
                        dep = LinkDependency(action['dev'], netns.netns)
                        if link in deps:
                            deps[link].append(dep)
                        else:
                            deps[link] = [dep]

        return deps

    def _stages(self, continue_on_circualar):
        def dep(arg):
            '''
                Dependency resolver

            "arg" is a dependency dictionary in which
            the values are the dependencies of their respective keys.
            '''
            d=dict((k, set(arg[k])) for k in arg)
            r=[]
            while d:
                # values not in keys (items without dep)
                t=set(i for v in d.values() for i in v)-set(d.keys())
                # and keys without value (items without dep)
                t.update(k for k, v in d.items() if not v)

                if len(t) == 0:
                    logger.error("Circualar link dependency detected: ")
                    for k, v in d.items():
                        logger.error('  {} => {}'.format(k, ", ".join(map(str, v))))
                    logger.error("")

                    if continue_on_circualar:
                        # remaining link deps cannot be resolved, drop them
                        # if we shall continue
                        return r
                    else:
                        raise LinkCircularLinked()

                # can be done right away
                r.append( sorted(t) )
                # and cleaned up
                d=dict(((k, v-t) for k, v in d.items() if v))
            return r

        dependencies = self._dependencies(self.root_netns)
        if self.namespaces is not None:
            for netns in self.namespaces.values():
                dependencies.update(self._dependencies(netns))

        if logger.getEffectiveLevel() <= logging.DEBUG:
            logger.debug('dependencies dump:')
            for ifname, deps in dependencies.items():
                logger.debug('  %s => %s', ifname, ", ".join(map(str, deps)))

        stages = dep(dependencies)

        if logger.getEffectiveLevel() <= logging.DEBUG:
            logger.debug('stages dump:')
            i = 1
            for stage in stages:
                logger.debug('  #%d => %s', i, ", ".join(map(str, stage)))
                i += 1

        return stages

    def _apply(self, do_apply, vrrp_type=None, vrrp_name=None, vrrp_state=None):
        # check if called from vrrp hook and ignore non-vrrp interfaces
        by_vrrp = not None in [
            vrrp_type, vrrp_name, vrrp_state]
        if by_vrrp:
            logger.info("triggered by vrrp state update")
            logger.log_change("{} {}".format(vrrp_type, vrrp_name), vrrp_state)
            logger.info("")

            # ifstate schema requires lower case keywords
            vrrp_type = vrrp_type.lower()
            vrrp_state = vrrp_state.lower()

        # create and destroy namespaces to match config
        if not by_vrrp and self.namespaces is not None:
            prepare_netns(do_apply, self.namespaces.keys(), self.new_namespaces)
            logger.info("")

        # get link dependency tree
        stages = self._stages(do_apply)

        # remove any orphan (non-ignored) links
        if not by_vrrp:
            had_cleanup = False
            cleanup_items = []
            for item in self.link_registry.registry:
                ifname = item.attributes['ifname']
                # items without a link are orphan - keep them if they match the ignore regex list...
                if item.link is None and not any(re.match(regex, ifname) for regex in self.ignore.get('ifname', [])):
                    # ...or are in a netns namespace while the config has no `namespaces` setting
                    if self.namespaces is not None or item.netns.netns is None:
                        if not had_cleanup:
                            logger.info("cleanup orphan interfaces...")
                            had_cleanup = True
                        if self.free_registry_item(do_apply, item):
                            cleanup_items.append(item)
                        else:
                            item.attributes['orphan'] = True

            if cleanup_items:
                for item in cleanup_items:
                    self.link_registry.registry.remove(item)

            if had_cleanup:
                logger.info("")

        # dump link registry in verbose mode
        if logger.getEffectiveLevel() <= logging.DEBUG:
            self.link_registry.debug_dump()

        if not by_vrrp:
            # apply bpf settings
            had_bpf = self._apply_bpf(do_apply, self.root_netns)
            if self.namespaces is not None:
                for name, netns in self.namespaces.items():
                    had_bpf = self._apply_bpf(do_apply, netns, had_bpf)
            if had_bpf:
                logger.info("")

            # apply sysctl settings
            had_sysctl = self._apply_sysctl(do_apply, self.root_netns)
            if self.namespaces is not None:
                for name, netns in self.namespaces.items():
                    had_sysctl = self._apply_sysctl(do_apply, netns, had_sysctl)
            if had_sysctl:
                logger.info("")

        # create/modify links in order of dependencies
        logger.info("configure interfaces...")
        for stage in stages:
            for link_dep in stage:
                if link_dep.netns is None:
                    self._apply_iface(do_apply, self.root_netns, link_dep, by_vrrp, vrrp_type, vrrp_name, vrrp_state)
                else:
                    self._apply_iface(do_apply, self.namespaces[link_dep.netns], link_dep, by_vrrp, vrrp_type, vrrp_name, vrrp_state)

        # configure routing
        logger.info("")
        logger.info("configure routing...")
        self._apply_routing(do_apply, self.root_netns, by_vrrp, vrrp_type, vrrp_name, vrrp_state)
        if self.namespaces is not None:
            for name, netns in self.namespaces.items():
                self._apply_routing(do_apply, netns, by_vrrp, vrrp_type, vrrp_name, vrrp_state)

    def _apply_bpf(self, do_apply, netns, had_bpf=False):
        if not netns.bpf_progs is None:
            if not had_bpf:
                logger.info("load BPF programs...")
                had_bpf = True
            netns.bpf_progs.apply(do_apply)

        return had_bpf

    def _apply_sysctl(self, do_apply, netns, had_sysctl=False):
        for iface in ['all', 'default']:
            if netns.sysctl.has_settings(iface):
                if not had_sysctl:
                    logger.info("configure sysctl settings...")
                    had_sysctl = True
                netns.sysctl.apply(iface, do_apply)

        if netns.sysctl.has_globals():
            if not had_sysctl:
                logger.info("configure sysctl settings...")
                had_sysctl = True
            netns.sysctl.apply_globals(do_apply)

        return had_sysctl

    def _apply_iface(self, do_apply, netns, link_dep, by_vrrp, vrrp_type, vrrp_name, vrrp_state):
        ifname = link_dep.ifname
        if ifname in netns.links:
            link = netns.links[ifname]

        # check for vrrp mode:
        #   disable: vrrp type & name matches, but vrrp state not
        #   ignore : vrrp type & name does not match
        if ifname in netns.vrrp['links']:
            # this is a vrrp link but not running in a vrrp action
            if not by_vrrp:
                return
            else:
                # skip if another vrrp type & name is addressed
                if not link.match_vrrp_select(vrrp_type, vrrp_name):
                    logger.debug('other vrrp',
                                extra={'iface': ifname})
                    return
                # vrrp type & name does match, but the state does not => disable this interface
                elif not vrrp_name in netns.vrrp[vrrp_type] or not vrrp_state in netns.vrrp[vrrp_type][vrrp_name] or not ifname in netns.vrrp[vrrp_type][vrrp_name][vrrp_state]:
                    if ifname in netns.links:
                        logger.debug('disabled due to vrrp constraint',
                                    extra={'iface': ifname})
                        link.settings['state'] = 'down'
        # ignore if this link is not vrrp aware at all
        elif by_vrrp:
            logger.debug('no vrrp setting',
                        extra={'iface': ifname})
            return

        logger.info(" {}".format(link_dep))

        if ifname in netns.links:
            excpts = link.apply(do_apply, netns.sysctl)
            if excpts.has_errno(errno.EEXIST):
                retry = True

        if ifname in netns.tc:
            netns.tc[ifname].apply(do_apply)

        if ifname in netns.xdp:
            netns.xdp[ifname].apply(do_apply, netns.bpf_progs)

        if ifname in netns.addresses and netns.addresses[ifname]:
            netns.addresses[ifname].apply(self.ipaddr_ignore, self.ignore.get(
                'ipaddr_dynamic', True), do_apply)

        if ifname in netns.fdb:
            netns.fdb[ifname].apply(do_apply)

        if ifname in netns.neighbours:
            netns.neighbours[ifname].apply(do_apply)

        if ifname in netns.wireguard:
            netns.wireguard[ifname].apply(do_apply)

    def _apply_routing(self, do_apply, netns, by_vrrp, vrrp_type, vrrp_name, vrrp_state):
        if not netns.tables is None:
            netns.tables.apply(self.ignore.get('routes', []), do_apply, by_vrrp, vrrp_type, vrrp_name, vrrp_state)

        if not netns.rules is None:
            netns.rules.apply(self.ignore.get('rules', []), do_apply, by_vrrp, vrrp_type, vrrp_name, vrrp_state)

    def show(self, showall=False):
        if showall:
            defaults = deepcopy(Parser._default_ifstates)
        else:
            defaults = {}

        ipaddr_ignore = []
        for ip in Parser._default_ifstates['ignore']['ipaddr_builtin']:
            ipaddr_ignore.append(ip_network(ip))

        root_config = self._show_netns(self.root_netns, showall, ipaddr_ignore)
        netns_instances = get_netns_instances()
        if len(netns_instances) > 0:
            netns_configs = {}
            for netns in netns_instances:
                netns_configs[netns.netns] = self._show_netns(netns, showall, ipaddr_ignore)

            return {**defaults, **root_config, 'namespaces': netns_configs}


        return {**defaults, **root_config}

    def _show_netns(self, netns, showall, ipaddr_ignore):
        ifs_links = []
        for ipr_link in netns.ipr.get_links():
            name = ipr_link.get_attr('IFLA_IFNAME')
            # skip links on ignore list
            if name != 'lo' and not any(re.match(regex, name) for regex in Parser._default_ifstates['ignore']['ifname_builtin']):
                ifs_link = {
                    'name': name,
                    'addresses': [],
                    'link': {
                        'state': ipr_link['state'],
                    },
                }

                for addr in netns.ipr.get_addr(index=ipr_link['index']):
                    if addr['flags'] & IFA_F_PERMANENT == IFA_F_PERMANENT:
                        ip = ip_interface(addr.get_attr(
                            'IFA_ADDRESS') + '/' + str(addr['prefixlen']))
                        if not any(ip in net for net in ipaddr_ignore):
                            ifs_link['addresses'].append(ip.with_prefixlen)

                info = ipr_link.get_attr('IFLA_LINKINFO')
                if info is None:
                    kind = None
                else:
                    kind = info.get_attr('IFLA_INFO_KIND')
                if kind is not None:
                    ifs_link['link']['kind'] = kind

                    data = info.get_attr('IFLA_INFO_DATA')
                    # unsupported link type, fallback to raw encoding
                    if data is not None and type(data) != str:
                        for k, v in data['attrs']:
                            if k not in ['UNKNOWN', 'IFLA_VLAN_FLAGS']:
                                attr = ipr_link.nla2name(k)
                                if attr in Link.attr_value_maps:
                                    ifs_link['link'][attr] = Link.attr_value_maps[attr].get(
                                        v, v)
                                elif attr in Link.attr_value_lookup:
                                    ifs_link['link'][attr] = Link.attr_value_lookup[attr].lookup_str(v)
                                else:
                                    ifs_link['link'][attr] = v
                else:
                    ifs_link['link']['kind'] = 'physical'
                    addr = ipr_link.get_attr('IFLA_ADDRESS')
                    if not addr is None:
                        ifs_link['link']['address'] = addr
                    permaddr = netns.ipr.get_permaddr(name)
                    if not permaddr is None:
                        if addr is None:
                            ifs_link['link']['addr'] = permaddr
                        elif addr != permaddr:
                            ifs_link['link']['permaddr'] = permaddr
                    businfo = netns.ipr.get_businfo(name)
                    if not businfo is None:
                        ifs_link['link']['businfo'] = businfo

                # add device group if not 0
                group = ipr_link.get_attr('IFLA_GROUP')
                if not group is None and group != 0:
                    ifs_link['link']['group'] = RTLookups.group.lookup_str(group)

                for attr in ['link', 'master', 'gre_link', 'ip6gre_link', 'vxlan_link', 'xfrm_link']:
                    ref = ipr_link.get_attr('IFLA_{}'.format(attr.upper()))
                    if ref is not None:
                        try:
                            ifs_link['link'][attr] = netns.ipr.get_ifname_by_index(
                                ref)
                        except Exception as err:
                            if not isinstance(err, netlinkerror_classes):
                                raise
                            logger.warning('lookup {} failed: {}'.format(
                                attr, err.args[1]), extra={'iface': name, 'netns': netns})
                            ifs_link['link'][attr] = ref

                mtu = ipr_link.get_attr('IFLA_MTU')
                if not mtu is None:
                    if not mtu in [1500, 65536] or name == 'lo':
                        ifs_link['link']['mtu'] = mtu

                brport.BRPort.show(netns.ipr, showall, ipr_link['index'], ifs_link)

                if name == 'lo':
                    if ifs_link['addresses'] == Parser._default_lo_link['addresses']:
                        del(ifs_link['addresses'])

                    if ifs_link['link'] == Parser._default_lo_link['link']:
                        del(ifs_link['link'])

                    if len(ifs_link) > 1:
                        ifs_links.append(ifs_link)
                else:
                    ifs_links.append(ifs_link)

        routing = {
            'routes': Tables(netns).show_routes(Parser._default_ifstates['ignore']['routes_builtin']),
            'rules': Rules(netns).show_rules(Parser._default_ifstates['ignore']['rules_builtin']),
        }

        return {**{'interfaces': ifs_links, 'routing': routing}}

    def get_defaults(self, **kwargs):
        for default in self.defaults:
            for match in default['match']:
                matching = True

                for option, regex in match.items():
                    if not option in kwargs:
                        matching = False
                    elif not re.match(regex, kwargs[option]):
                        matching = False

                if matching:
                    return default

        return {}

    def gen_unique_ifname(self):
        '''
        Get a random unique ifname over all namespaces and configured ifnames.
        '''
        while True:
            ifname = "ifs.tmp.{}".format(secrets.token_hex(3))
            item = self.link_registry.get_link(ifname=ifname)
            if item is None:
                return ifname



from libifstate.util import logger, IfStateLogging
from libifstate.exception import netlinkerror_classes
from libifstate.bpf import libbpf, struct_bpf_prog_info, bpfs_ifstate_dir

from pyroute2 import IPRoute

from pyroute2.netlink.rtnl.ifinfmsg import XDP_FLAGS_SKB_MODE
from pyroute2.netlink.rtnl.ifinfmsg import XDP_FLAGS_DRV_MODE
from pyroute2.netlink.rtnl.ifinfmsg import XDP_FLAGS_HW_MODE

import ctypes
import os
import pyroute2.netns
import shutil

class XDP():
    def __init__(self, netns, iface, xdp):
        self.netns = netns
        self.iface = iface
        self.xdp = xdp

    def apply(self, do_apply, bpf_progs):
        self.link = next(iter(self.netns.ipr.get_links(ifname=self.iface)), None)

        if self.link == None:
            logger.warning('link missing', extra={'iface': self.iface})
            return

        if self.netns.netns is not None:
            pyroute2.netns.pushns(self.netns.netns)

        try:
            # get ifindex
            self.idx = self.link['index']

            # new BPF instance (from pinned, bpf or xdp settings)
            new_prog_tag = None
            new_prog_fd = None

            # get current BPF tag, if any
            current_prog_tag = None
            current_prog_id = None
            current_attached = None
            ifla_xdp = self.link.get_attr('IFLA_XDP')
            if ifla_xdp is not None:
                current_prog_id = ifla_xdp.get_attr('IFLA_XDP_PROG_ID')
                current_attached = ifla_xdp.get_attr('IFLA_XDP_ATTACHED')

            if current_prog_id:
                current_prog_fd = libbpf.bpf_prog_get_fd_by_id(current_prog_id)
                if current_prog_fd < 0:
                    logger.warning('could not get current XDP prog fd for {}: {}'.format(
                        self.iface, os.strerror(-current_prog_fd)))
                else:
                    current_prog_info = struct_bpf_prog_info()
                    current_prog_size = ctypes.c_uint(ctypes.sizeof(current_prog_info))
                    rc = libbpf.bpf_obj_get_info_by_fd(current_prog_fd, ctypes.byref(current_prog_info), ctypes.byref(current_prog_size))
                    if rc == 0:
                        current_prog_tag = bytes(current_prog_info.tag).hex()

                        logger.debug('current prog tag: {}'.format(current_prog_tag), extra={
                                    'iface': self.iface})
                    else:
                        logger.warning('could not get current XDP obj info for {}: {}'.format(
                            self.iface, os.strerror(-rc)))

            logger.debug('current attached: {}'.format(current_attached), extra={
                'iface': self.iface})

            # load new BPF prog
            new_obj = None
            if not self.xdp:
                new_prog_fd = -1
            elif 'pinned' in self.xdp:
                new_prog_fd = libbpf.bpf_obj_get( os.fsencode(self.xdp["pinned"]) )

                if new_prog_fd < 0:
                    logger.warning("pinned BPF object '{}' cannot be opened: {}".format(
                        self.xdp["pinned"], os.strerror(-new_prog_fd)))
            elif 'bpf' in self.xdp:
                if bpf_progs is not None:
                    (new_prog_fd, new_prog_tag) = bpf_progs.get_bpf(self.xdp["bpf"])

                if new_prog_fd == -1 or new_prog_fd is None:
                    logger.warning("BPF program '{}' not loaded for {}".format(
                        self.xdp["bpf"], self.iface))
                else:
                    logger.debug('new prog tag: {}'.format(new_prog_tag), extra={
                                'iface': self.iface})
            else:
                assert False, "unhandled XDP settings"

            # get new prog tag
            if new_prog_fd != -1 and new_prog_fd != None and new_prog_tag == None:
                info = struct_bpf_prog_info()
                sz = ctypes.c_uint(ctypes.sizeof(info))
                rc = libbpf.bpf_obj_get_info_by_fd(new_prog_fd, ctypes.byref(info), ctypes.byref(sz))
                if rc == 0:
                    new_prog_tag = bytes(info.tag).hex()

                    logger.debug('new prog tag: {}'.format(new_prog_tag), extra={
                                'iface': self.iface})
                else:
                    logger.warning('XDP failed to get prog info on {}'.format(
                        self.iface))

            # get attach mode flags
            new_attached = self.xdp.get("mode", "auto")
            if type(new_attached) != list:
                new_attached = [new_attached]

            new_flags = 0
            if "auto" in new_attached:
                new_attached = ["xdp", "xdpgeneric", "xdpoffload"]
            else:
                if "xdp" in new_attached:
                    new_flags = new_flags | XDP_FLAGS_DRV_MODE
                if "xdpgeneric" in new_attached:
                    new_flags = new_flags | XDP_FLAGS_SKB_MODE
                if "xdpoffload" in new_attached:
                    new_flags = new_flags | XDP_FLAGS_HW_MODE

            logger.debug('new attached: {}'.format(", ".join(new_attached)), extra={
                'iface': self.iface})

            logger.debug('new flags: {}'.format(new_flags), extra={
                'iface': self.iface})

            # set new XDP prog if tag or attach mode has changed
            if current_prog_tag != new_prog_tag or not current_attached in new_attached:
                if do_apply:
                    ipr = IPRoute()
                    attach_ok = False
                    try:
                        ipr.link('set', index=self.idx, xdp_fd=new_prog_fd, xdp_flags=new_flags)
                        attach_ok = True
                    except Exception as err:
                        if not isinstance(err, netlinkerror_classes):
                            raise

                        # try again: detaching XDP first
                        try:
                            ipr.link('set', index=self.idx, xdp_fd=-1)

                            ipr.link('set', index=self.idx, xdp_fd=new_prog_fd, xdp_flags=new_flags)
                            attach_ok = True
                        except Exception as err:
                            if not isinstance(err, netlinkerror_classes):
                                raise

                            logger.warning('attaching XDP program on {} failed: {}'.format(
                                self.iface, err.args[1]))

                if new_prog_fd == -1:
                    logger.log_del('xdp', 'detach')
                else:
                    logger.log_change('xdp')
            else:
                logger.log_ok('xdp')

        finally:
            if self.netns.netns is not None:
                pyroute2.netns.popns()

try:
    libbpf.bpf_obj_get_info_by_fd
    libbpf.bpf_object__name
    libbpf.bpf_object__next_program
    libbpf.bpf_object__load
    libbpf.bpf_object__open_file
    libbpf.bpf_object__pin_maps
    libbpf.bpf_prog_get_fd_by_id
    libbpf.bpf_program__fd
    libbpf.bpf_program__section_name
except AttributeError as ex:
    # ignore missing library
    raise ModuleNotFoundError(ex)



from libifstate.util import logger, IfStateLogging
from libifstate.exception import netlinkerror_classes, FeatureMissingError
from wgnlpy import WireGuard as WG
from ipaddress import ip_network
import collections
from copy import deepcopy
import pyroute2.netns
import socket

class WireGuard():
    def __init__(self, netns, iface, wireguard):
        self.netns = netns
        self.iface = iface
        self.wireguard = wireguard

        if self.netns.netns is not None:
            pyroute2.netns.pushns(self.netns.netns)

        try:
            self.wg = WG()
        finally:
            if self.netns.netns is not None:
                pyroute2.netns.popns()

        # convert allowedips peers settings into IPv[46]Network objects
        # and remove duplicates
        if 'peers' in self.wireguard:
            for i, peer in enumerate(self.wireguard['peers']):
                if 'allowedips' in peer:
                    self.wireguard['peers'][i]['allowedips'] = set(
                        [ip_network(x) for x in self.wireguard['peers'][i]['allowedips']])

    def __deepcopy__(self, memo):
        '''
        Add custom deepcopy implementation to keep single wgnlpy.WireGuard instances.
        '''
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if k == 'wg':
                if self.netns.netns is not None:
                    pyroute2.netns.pushns(self.netns.netns)

                try:
                    setattr(result, k, WG())
                finally:
                    if self.netns.netns is not None:
                        pyroute2.netns.popns()
            else:
                setattr(result, k, deepcopy(v, memo))
        return result

    def apply(self, do_apply):
        # get kernel's wireguard settings for the interface
        try:
            state = self.wg.get_interface(
                self.iface, spill_private_key=True, spill_preshared_keys=True)
        except Exception as err:
            logger.warning('WireGuard on {} failed: {}'.format(
                self.iface, err.args[1]))
            return

        # check base settings (not the peers, yet)
        has_changes = False
        for setting in [x for x in self.wireguard.keys() if x != 'peers']:
            logger.debug('  %s: %s => %s', setting, getattr(
                state, setting), self.wireguard[setting], extra={'iface': self.iface})
            has_changes |= self.wireguard[setting] != getattr(state, setting)

        if has_changes:
            logger.log_change('wireguard')
            if do_apply:
                try:
                    self.wg.set_interface(
                        self.iface, **{k: v for k, v in self.wireguard.items() if k != "peers"})
                except Exception as err:
                    if not isinstance(err, netlinkerror_classes):
                        raise
                    logger.warning('updating iface {} failed: {}'.format(
                        self.iface, err.args[1]))
        else:
            logger.log_ok('wireguard')

        # check peers list if provided
        if 'peers' in self.wireguard:
            peers = getattr(state, 'peers')
            has_pchanges = False

            avail = []
            for peer in self.wireguard['peers']:
                avail.append(peer['public_key'])
                pubkey = next(
                    iter([x for x in peers.keys() if x == peer['public_key']]), None)
                if pubkey is None:
                    has_pchanges = True
                    if do_apply:
                        try:
                            self.safe_set_peer(peer)
                        except Exception as err:
                            if not isinstance(err, netlinkerror_classes):
                                raise
                            logger.warning('add peer to {} failed: {}'.format(
                                self.iface, err.args[1]))
                else:
                    pchange = False
                    for setting in peer.keys():
                        attr = getattr(peers[pubkey], setting)
                        if setting == 'allowedips':
                            attr = set(attr)
                        logger.debug('  peer.%s: %s => %s', setting, attr,
                                     peer[setting], extra={'iface': self.iface})
                        if type(attr) == set:
                            pchange |= not (attr == peer[setting])
                        else:
                            pchange |= str(peer[setting]) != str(getattr(
                                peers[pubkey], setting))

                    if pchange:
                        has_pchanges = True
                        if do_apply:
                            try:
                                self.safe_set_peer(peer)
                            except Exception as err:
                                if not isinstance(err, netlinkerror_classes):
                                    raise
                                logger.warning('change peer at {} failed: {}'.format(
                                    self.iface, err.args[1]))

            for peer in peers:
                if not peer in avail:
                    has_pchanges = True
                    if do_apply:
                        try:
                            self.wg.remove_peers(self.iface, peer)
                        except Exception as err:
                            if not isinstance(err, netlinkerror_classes):
                                raise
                            logger.warning('remove peer from {} failed: {}'.format(
                                self.iface, err.args[1]))
            if has_pchanges:
                logger.log_change('wg.peers')
            else:
                logger.log_ok('wg.peers')

    def safe_set_peer(self, peer):
        try:
            self.wg.set_peer(self.iface, **peer)
        except (socket.gaierror, ValueError) as err:
            logger.warning('failed to set wireguard endpoint at {}: {}'.format(self.iface, err))

            del(peer['endpoint'])
            self.wg.set_peer(self.iface, **peer)



from libifstate.util import logger
from libifstate.parser.base import Parser
from libifstate.exception import ParserOpenError, ParserParseError, ParserIncludeError
import yaml
import os


class Loader(yaml.SafeLoader):
    def __init__(self, stream):
        try:
            self._basedir = os.path.split(stream.name)[0]
        except AttributeError:
            pass

        if not self._basedir:
            self._basedir = os.path.curdir

        super(Loader, self).__init__(stream)
        self.add_constructor('!include', self.include)

    def include(self, id, node):
        filename = self.construct_scalar(node)
        if not os.path.isabs(filename):
            filename = os.path.join(self._basedir, filename)

        try:
            with open(filename, 'r') as fh:
                return yaml.load(fh, Loader)
        except OSError as ex:
            raise ParserIncludeError(ex)
        except yaml.parser.ParserError as ex:
            raise ParserParseError(ex)


class YamlParser(Parser):
    def __init__(self, fn):
        logger.debug('YamlParser parsing %s', fn)
        try:
            with open(fn) as fh:
                self.ifstates = yaml.load(fh, Loader)
        except OSError as ex:
            raise ParserOpenError(ex)
        except yaml.parser.ParserError as ex:
            raise ParserParseError(ex)
